"""

Compute statistics for:
    - preprocessing: mean & std for images and labels (normalization)
    - prediction: summarize cloud-related info (uncertainty approximation)

"""

import argparse
from blowtorch import Run
import h5py
import pickle
import os
import numpy as np
import yaml
from pathlib import Path

from src.utils.load_s2_and_target import read_sentinel2_bands, save_array_as_geotif

def config():
    a = argparse.ArgumentParser()
    # stats for preprocessing (default)
    a.add_argument("--preproconfig", help="path to data config", default='configs/train.yaml')

    # stats for prediction (e.g. cloud-related stats, use configs/predictSet.yaml)
    a.add_argument("--predstat", default=False, type=bool, help="whether to summarize cloud statistics",)
    a.add_argument("--predconfig", help="path to prediction config", default='configs/predictSet.yaml')

    args = a.parse_args()
    return args

def compute_stats_cloud(cfgs, cloud_thresh):
    '''
    Compute stats for prediction (# cloud-free pixels & avg. cloud prob. )
    :param
    cfgs: e.g. configs/predictSet.yaml
    cloud_thresh: e.g. 10, (cloud_prob<=cloud_thresh => cloud-free)
    :return:
    '''
    for year in cfgs['years']:
        for tile in cfgs['tiles']:
            print("PROCECESSING YEAR{} - TILE{}".format(year, tile))

            s2_dir = os.path.join(cfgs['path_cfg']['s2_dir'], 'CH_' + year, 'sentinel_2A')
            img_paths = [img_path for img_path in os.listdir(s2_dir) if tile in img_path]

            cloudfree_num, cloudprob_sum = None, None
            validimg_num = 0
            tile_info = None
            for imgidx, img_path in enumerate(img_paths):
                # read cloud_mask from image
                _, cloud_array = read_sentinel2_bands(data_path=os.path.join(s2_dir, img_path), channels_last=False)

                # check invalid image
                if cloud_array is None:
                    continue

                # summarize cloud-related info
                validimg_num+=1
                if cloudfree_num is None:
                    cloudfree_num = np.zeros(cloud_array.shape, dtype=np.int8)
                    cloudprob_sum = np.zeros(cloud_array.shape)
                    tile_info = read_sentinel2_bands(data_path=os.path.join(s2_dir, img_path), channels_last=False, info_only=True)

                cloudfree_num += (cloud_array <= cloud_thresh)
                cloudprob_sum += cloud_array

            print('NUM VALID IMAGE: {}'.format(validimg_num))
            # save avg. cloud_prob
            cloudprob_sum /= validimg_num
            out_path = Path(cfgs['path_cfg']['out_dir']) / 'clouds' / (year + tile + '_avgprob.tif')
            out_path.parent.mkdir(parents=True, exist_ok=True)
            save_array_as_geotif(str(out_path), np.float32(cloudprob_sum), tile_info)

            # save num of cloud_free pixels
            out_path = Path(cfgs['path_cfg']['out_dir']) / 'clouds' / (year + tile + '_num.tif')
            out_path.parent.mkdir(parents=True, exist_ok=True)
            save_array_as_geotif(str(out_path), np.float32(cloudfree_num), tile_info)

def compute_stats(cfgs):
    '''
    Compute stats for preprocessing (mean & std)
    :param cfgs: e.g. train.yaml
    :return:
    '''
    img_sums = []
    img_squared_sums = []
    label_sums = []
    label_squared_sums = []

    num_image_valid = 0
    num_label_valid = 0
    # read each tile
    for tile in cfgs['tiles']:

        # get mask
        pkl_file = os.path.join(cfgs['h5_dir'], '{}_train_val_mask.pkl'.format(tile))
        with open(pkl_file, 'rb') as f:
            data = pickle.load(f)
        valid_location = data['valid_location']

        # get data
        h5_file = os.path.join(cfgs['h5_dir'], '{}.h5'.format(tile))
        hf = h5py.File(h5_file, 'r')

        data = hf.get('data')
        image = data.get('images') # imgid, C, H, W
        orig_labels = data.get('labels')
        # organize label shape
        width, height = int(orig_labels.shape[0] / 2), orig_labels.shape[1]
        label = np.stack((orig_labels[:width, :], orig_labels[width:, :])) # 2, H, W
        label = np.expand_dims(label, axis=0)

        image = (image * valid_location).astype(np.int32)
        label = np.nan_to_num(label) * valid_location
        
        print('negative count {}, nan count {}'.format(np.sum(label<0), np.sum(np.isnan(label))))
        img_sums.append(np.sum(image, axis=(0, 2, 3)))
        img_squared_sums.append(np.sum(image**2, axis=(0, 2, 3)))
        label_sums.append(np.sum(label, axis=(0, 2, 3)))
        label_squared_sums.append(np.sum(label**2, axis=(0, 2, 3)))
        
        num_image_valid += valid_location.sum()*image.shape[0]
        num_label_valid += valid_location.sum()*label.shape[0]

        hf.close()

    img_mean, img_std = getMeanStd(np.array(img_sums), np.array(img_squared_sums), num_image_valid)
    label_mean, label_std = getMeanStd(np.array(label_sums), np.array(label_squared_sums), num_label_valid)

    print(img_mean, img_std, label_mean, label_std)

    # save data stat for each tile
    with open(os.path.join(cfgs['h5_dir'], 'stats_total.yaml'), 'w') as fh:
        yaml.dump({'img_mean': img_mean, 'img_std': img_std, 'label_mean': label_mean, 'label_std': label_std}, fh)

def getMeanStd(sums, sqaured_sums, num_valid):
    mean = np.sum(sums, axis=0)/float(num_valid)
    sqaured_mean = np.sum(sqaured_sums, axis=0)/float(num_valid)
    std = np.sqrt(sqaured_mean - np.square(mean))
    return mean.tolist(), std.tolist()

if __name__ == "__main__":

    args = config()

    if args.predstat:
        run = Run(config_files=[args.predconfig])
        run.seed_all(12345)
        with open(args.preproconfig) as fh:
            prepro_config = yaml.safe_load(fh)
        cloud_thresh = prepro_config['cloud_thresh']

        compute_stats_cloud(run, cloud_thresh)
    else:
        run = Run(config_files=[args.preproconfig])
        run.seed_all(12345)
        compute_stats(run['data_cfg'])


