"""
process images data & reference data into pkl file
"""
import argparse
from blowtorch import Run
import random
import os
import numpy as np
from collections import defaultdict
import pickle
from tqdm import trange
import rasterio
import fiona
import rasterio.warp
import rasterio.features
import yaml
from scipy.signal import convolve2d
import h5py

from src.utils.load_s2_and_target import load_tif_as_array, read_sentinel2_bands

def config():
    a = argparse.ArgumentParser()
    a.add_argument("--config", help="path to preprocess config", default='configs/preprocess_euler.yaml')
    args = a.parse_args()
    return args

def prepareData(cfgs):
    '''

    :param cfgs: check preprocess.yaml
    :return: save h5 files {data: {images, labels}, train: {locations, loc2img}, val: {locations, loc2img}}
    {locations, loc2img}: location ([i, j]), loc2img: (imgid, i, j)
    '''
    print('OUTPUT_DIR: {}'.format(cfgs['output_dir']))
    
    for tid in cfgs['tiles']:

        valid_location_mask = None
        valid_image_mask = []

        print('== PROCESSING tid {}'.format(tid))
        ##################################################
        #                load target                     #
        ##################################################
        labels = None
        valid_target_mask = None
        for v in cfgs['variables']:
            target_file = os.path.join(cfgs['target_dir'], 'reprojected_10m_{}'.format(v),
                                       'mosaic_chm_2017_2018_2019_reprojected_10m_{}_{}.tif'.format(v, tid))
            target, tile_info = load_tif_as_array(target_file)
            target = target.astype(np.float32)
            # mask negative values (nodata) to nan
            target[target < 0] = np.nan
            if labels is None:
                labels = target
                valid_target_mask = ~np.isnan(target)
            else:
                labels = np.vstack((labels, target))
                # valid in all variables targets
                valid_target_mask = valid_target_mask * (~np.isnan(target))
        
        # skip borders
        patch_half = cfgs['patch_size'] // 2
        valid_target_mask[:patch_half, :] = 0
        valid_target_mask[-patch_half:, :] = 0
        valid_target_mask[:, :patch_half] = 0
        valid_target_mask[:, -patch_half:] = 0

        # get train & val split mask  (0 -> train, 1-> val)
        split_mask = splitTrainVal(cfgs['trainvalshp'][0], cfgs['trainvalshp'][1], target_file)

        # remove overlap patches around train/val boundary
        k = np.ones((cfgs['patch_size'], cfgs['patch_size']), dtype=int)
        new = convolve2d(split_mask, k, 'same', 'symm')
        split_mask = np.where((new > 0) & (new < cfgs['patch_size'] * cfgs['patch_size']), 2, split_mask)

        # valid target = valid target & inside area of interest
        valid_target_mask = valid_target_mask * (projectToTarget(cfgs['trainvalshp'][2], rasterio.open(target_file)) == 1)

        ##################################################
        #                load image                      #
        ##################################################
        images = []
        imgid = 0

        for y in cfgs['years']:

            year_mask = np.ones(split_mask.shape, dtype=split_mask.dtype)
            # check captured-year of reference data
            if cfgs['split_year']:
                year_file = os.path.join(cfgs['yearshp_dir'], y+'.shp')
                if not os.path.exists(year_file):
                    print('NO REFERENCE DATA FOR YEAR {}'.format(y))
                    continue
                # year_mask = 1 for valid data, = 0 for no data
                year_mask = getYearMask(year_file, target_file)

                # if no reference info for this tile this year, skip
                if np.sum(year_mask) == 0:
                    continue

            # find images (num_image images, total num img = num_date * num_orbit)
            orbits = {}
            dates = []
            orbitcount, imgcount = 0, 0
            # load image in each year and save
            image_dir = os.path.join(cfgs['image_dir'], 'CH_'+y, 'sentinel_2A')
            if not os.path.exists(image_dir):
                print('== NOT FOUND YEAR {}'.format(y))
                continue
            zipfiles = os.listdir(image_dir)
            if len(zipfiles) < 1:
                print('== NOT FOUND YEAR {}'.format(y))
                continue
            for zipfile in zipfiles:
                useimage = False
                image_array, cloud_array = None, None
                image_path = os.path.join(image_dir, zipfile)
                # get date, orbit info
                date, orbit, tile, _ = getAcqInfo(zipfile)

                # check tile id
                if not tid == tile:
                    continue

                if orbit in list(orbits.keys()):
                    if date not in orbits[orbit] and date not in dates and len(orbits[orbit]) < cfgs['num_date']:
                        print(' {} not in orbits[{}], len(orbits[orbit]) is {}'.format(date, orbit, len(orbits[orbit])))
                        orbits[orbit].append(date)
                        dates.append(date)
                        useimage = True
                else:
                    if orbitcount < cfgs['num_orbit'] and date not in dates:
                        print('add new orbit {}'.format(orbit))
                        orbits[orbit] = []
                        orbits[orbit].append(date)
                        dates.append(date)
                        orbitcount += 1
                        useimage = True
                if useimage:
                    image_array, cloud_array = read_sentinel2_bands(data_path=image_path, channels_last=False)
                    if image_array is None or cloud_array is None:
                        print('IMAGE CANNOT BE LOADED: {}'.format(image_path))
                        continue

                    print('== USE IMAGE {}, ORBITCOUNT {}, IMGCOUNT {}'.format(image_path, orbitcount, imgcount))
                    images.append(image_array)

                    cloudfree_mask = cloud_array <= cfgs['cloud_thresh']
                    nonempty_mask = np.sum(image_array[1:4, :, :], axis=0) != 0

                    valid_mask_final = valid_target_mask * cloudfree_mask * nonempty_mask * year_mask
                    valid_image_mask.append(valid_mask_final.astype(np.uint8)*(imgid+1))

                    if valid_location_mask is None:
                        valid_location_mask = valid_mask_final
                    else:
                        valid_location_mask += valid_mask_final
                    
                    imgcount += 1
                    imgid += 1

                    print('== FINISH LOAD IMAGE {}'.format(image_path))
                else:
                    print('== -SKIP IMAGE {}'.format(image_path))
                    continue

                if imgcount >= cfgs['num_orbit']*cfgs['num_date']:
                    break

        # no images corresponde to this reference data
        if len(images) < 1:
            print('== NO IMAGES FOR {}'.format(target_file))
            continue
        print('== START FIND INDEX FOR {} IMAGES'.format(len(images)))

        # dump to pickle (for visualization)
        if not os.path.exists(cfgs['output_dir']):
            os.mkdir(cfgs['output_dir'])
        output_file = os.path.join(cfgs['output_dir'], '{}_train_val_mask.pkl'.format(tid))
        with open(output_file, 'wb') as f:
            pickle.dump({
                'valid_location': valid_location_mask,
                'split_mask': split_mask,
            }, f, protocol=pickle.HIGHEST_PROTOCOL)
        print('== SAVE TO '+output_file)

        ##################################################
        #                find index                      #
        ##################################################

        locations = defaultdict()


        valid_location_mask[valid_location_mask > 0] = 1

        locations['train'] = np.argwhere(valid_location_mask * (split_mask == 0))
        locations['val'] = np.argwhere(valid_location_mask * (split_mask == 1))
        
        print('sum of year mask {}, sum of nonzero valid mask {}'.format(np.sum(year_mask), np.sum(valid_location_mask)))
        print('len of train locations {}'.format(len(locations['train'])))

        # dump to hdf5
        if not os.path.exists(cfgs['output_dir']):
            os.mkdir(cfgs['output_dir'])
        output_file = os.path.join(cfgs['output_dir'], '{}.h5'.format(tid))

        hf = h5py.File(output_file, 'w')
        data = hf.create_group('data')
        data.create_dataset('labels', data=labels)
        data.create_dataset('images', data=images)
        data.create_dataset('valid_image_mask', data=valid_image_mask)

        train = hf.create_group('train')
        train.create_dataset('locations', data=locations['train'])

        val = hf.create_group('val')
        val.create_dataset('locations', data=locations['val'])

        hf.close()

        print('== SAVE TO '+output_file)

        # check num_train, num_val for each tile
        num_train, num_val = len(locations['train']), len(locations['val'])
        print('== STAT, NUM_TRAIN {}, NUM_VAL {}'.format(num_train, num_val))

        # save data stat for each tile
        with open(os.path.join(cfgs['log_dir'], 'stats_{}.yaml'.format(tid)), 'w') as fh:
            yaml.dump({'num_train': num_train, 'num_val': num_val}, fh)

'''
get split mask based on train val shp
output array with target_file shape (0 -> train, 1-> val)
'''
def splitTrainVal(trainshp, valshp, target_file):
    target = rasterio.open(target_file)
    trainmask = projectToTarget(trainshp, target)
    valmask = projectToTarget(valshp, target)

    mask = np.ones(target.shape).astype(int)*2
    mask[trainmask==1] = 0
    mask[valmask==1] = 1
    return mask

'''
get year mask (with reference data captured in this year = 1, otw = 0)
'''
def getYearMask(yearshp_file, target_file):
    target = rasterio.open(target_file)
    yearmask = projectToTarget(yearshp_file, target)

    mask = np.ones(target.shape).astype(int)
    mask[yearmask==1] = 1
    mask[yearmask==0] = 0

    return yearmask

'''
project src shapefile to target crs with target.shape
'''
def projectToTarget(srcshp, target):
    polygon = fiona.open(srcshp)
    srccrs = polygon.crs
    poly = polygon.next()

    poly_trans = rasterio.warp.transform_geom(srccrs, target.crs, poly['geometry'])
    rasterized_polygon = rasterio.features.rasterize(
        [(poly_trans, 1)],
        out_shape=target.shape,
        transform=target.transform,
        fill=0,
        dtype='uint8'
    )
    return rasterized_polygon

'''
Get acquisition info of s2 image
e.g. name: S2A_MSIL2A_20170510T103031_N0205_R108_T31TGN_20170510T103025.zip
'''
def getAcqInfo(filename):
    infos = filename.split('.')[0].split('_')
    date = infos[2][:8]
    time = infos[2][8:]
    orbit = infos[4]
    tile = infos[5][1:]

    return date, orbit, tile, time


if __name__ == "__main__":

    args = config()
    run = Run(config_files=[args.config])
    run.seed_all(12345)

    if not os.path.exists(run['log_dir']):
        os.mkdir(run['log_dir'])
    if not os.path.exists(run['output_dir']):
        os.mkdir(run['output_dir'])

    prepareData(run)

    # save config file along with the pickle objects
    with open(os.path.join(run['log_dir'], 'data_config.yaml'), 'w') as f:
        yaml.dump(run.get_raw_config(), f)    
