"""
Dataset for canopy height estimation (CH)
"""

import random
import numpy as np
import os
import h5py
import yaml
from typing import List
from collections import defaultdict
import torch
from torch import Tensor
from torch.utils.data import Dataset, DataLoader, Sampler
from torchvision.transforms import Compose, Normalize

class HeightData():
    """
    Provides DataLoader objects for the pickled CH_Height dataset.
    """
    def __init__(self,
                 h5_dir: str,
                 subsample: float,
                 tiles: List[str],
                 patch_size: int,
                 normalize_img: bool,
                 normalize_labels: bool,
                 loader_args: dict,
                 iter_train: None, iter_val: None):

        images = None # imageid, c, h, w
        labels = None # tid, 2, h, w
        locations = defaultdict(list)
        valid_image_masks = None
        #  TODO: not hardcode num_img of tile of split year mode

        # 1 image per orbit
        num_img_tile = {'32TMS':6, '32TMT':4, '32TLT':2, '32TNS':4, '32TNT':2, '32TLS':4}
        # 4 image per orbit
        num_img_tile = {'32TMS':23, '32TMT':16, '32TLT':8, '32TNS':16, '32TNT':8, '32TLS':16}

        idx_start_tile = {}
        idx_start = 0
        for tid, tile in enumerate(tiles):
            print('=== process {}'.format(tile))
            h5_file = os.path.join(h5_dir, '{}.h5'.format(tile))
            hf = h5py.File(h5_file, 'r')

            # get image and label data
            data = hf.get('data')
            image = data.get('images')
            orig_labels = data.get('labels')
            # organize label shape
            width = int(orig_labels.shape[0]/2)
            label = np.zeros((1, 2, width, width)).astype(np.int16)
            label[0, 0, :, :] = orig_labels[:width, :]
            label[0, 1, :, :] = orig_labels[width:, :]
            #label = np.stack((orig_labels[:width, :], orig_labels[width:, :]))
            #label = np.expand_dims(label, axis=0)

            valid_image_mask = data.get('valid_image_mask')
            #print('sum valid_image_mask {}'.format(np.array(valid_image_mask).sum()))
            # save image and label
            if images is None:
                # TODO: not hardcode num_img_total of split year mode
                num_img_total = 87  if h5_dir.split('_')[-1] == 'splityear' else len(image)*len(tiles)
                print('== total num_img_total: {}'.format(num_img_total))
                images = np.zeros((num_img_total, image.shape[-3], image.shape[-2], image.shape[-1])).astype(np.int16)
                labels = np.zeros((len(tiles), label.shape[-3], label.shape[-2], label.shape[-1])).astype(np.int16)
                valid_image_masks = np.zeros((num_img_total, image.shape[-2], image.shape[-1])).astype(np.uint8)
            
            idx_end = idx_start+num_img_tile[tile] if h5_dir.split('_')[-1] == 'splityear' else len(image)*(tid+1)
            images[idx_start:idx_end, :, :, :] = image
            labels[tid:tid+1, :, :, :] = label
            valid_image_masks[idx_start:idx_end, :, :] = valid_image_mask
            
            idx_start_tile[tile] = idx_start
            idx_start = idx_end

            # get locations and loc2img
            for s in ['train', 'val']:
                location_info = hf.get(s)

                # skip empty data (e.g. no validation points in some tiles)
                if len(location_info.get('locations')) < 1:
                    continue

                if len(locations[s]) < 1:
                    locations[s] = np.zeros((len(location_info.get('locations')), 3)).astype(np.uint16)
                    locations[s][:, 0] = np.ones((len(location_info.get('locations')),))*tid
                    locations[s][:, 1:] = location_info.get('locations')
                    #locations[s] = np.hstack((np.ones((len(location_info.get('locations')), 1))*tid, location_info.get('locations'))).astype(np.uint16)
                else:
                    locations[s] = np.vstack((locations[s], np.hstack((np.ones((len(location_info.get('locations')), 1))*tid, location_info.get('locations'))))).astype(np.uint16)

                # filter out border points
                idx = np.argwhere(~((locations[s][:, 1]<7) | (locations[s][:, 1]>label.shape[-1]-7) | (locations[s][:, 2]<7) |  (locations[s][:, 2]>label.shape[-1]-7) ))
                locations[s] = np.squeeze(locations[s][idx])

            hf.close()

        print('images shape {}, labels shape {}'.format(images.shape, labels.shape))

        print('=== START BUILDING TRAIN SET')

        # stats for normalization
        with open(os.path.join(h5_dir, 'stats_total.yaml'), 'r') as f:
            stats = yaml.full_load(f)
        image_mean, image_std = stats['img_mean'], stats['img_std']
        label_mean, label_std = stats['label_mean'], stats['label_std']

        if normalize_img:
            image_trans = Compose([ToTensor(), Normalize(image_mean, image_std)])
        else:
            image_trans = Compose([ToTensor()])

        if normalize_labels:
            label_trans = Compose([ToTensor(), Normalize(label_mean, label_std)])
        else:
            label_trans = Compose([ToTensor()])
        
        print('norm pre ok')

        # creat train & val datasets
        train_set = _HeightDataset(images, labels, locations['train'], valid_image_masks, patch_size,
                                              image_transform=image_trans, label_transform=label_trans, 
                                              tile_info=[tiles, num_img_tile, idx_start_tile] if h5_dir.split('_')[-1] == 'splityear' else None)
        
        print('== FINISH BUILDING TRAIN SET')
        print('== START BUILDING VAL SET')

        val_set = _HeightDataset(images, labels, locations['val'], valid_image_masks, patch_size,
                                 image_transform=image_trans, label_transform=label_trans, 
                                 tile_info=[tiles, num_img_tile, idx_start_tile] if h5_dir.split('_')[-1] == 'splityear' else None)

        print('== CREATE TRAIN SET: LEN({}), VAL SET: LEN({})'.format(len(train_set), len(val_set)))

        # create train & val sub-samplers
        if iter_train:
            train_sampler = RandomSampler(train_set, num_samples=min(iter_train*loader_args['batch_size'], len(train_set))) if len(train_set)>0 else None
            val_sampler = RandomSampler(val_set, num_samples=min(iter_val*loader_args['batch_size'], len(val_set))) if len(val_set)>0 else None
            shuffle = False
        elif subsample and subsample < 1.0:
            train_sampler = RandomSampler(train_set, num_samples=int(subsample * len(train_set)))
            val_sampler = RandomSampler(val_set, num_samples=int(subsample * len(val_set)))
            shuffle = False
        else:
            train_sampler, val_sampler, shuffle = None, None, True

        # create train & val dataloaders
        self.train_loader = DataLoader(train_set, shuffle=shuffle, sampler=train_sampler, **loader_args)
        self.val_loader = DataLoader(val_set, shuffle=False, sampler=val_sampler, **loader_args)


class _HeightDataset(Dataset):
    def __init__(self,
                 images: np.array,
                 labels: np.array,
                 locations: np.array,
                 valid_image_masks: np.array,
                 patch_size: int,
                 image_transform: torch.nn.Module = None,
                 label_transform: torch.nn.Module = None, 
                 tile_info: list = None):

        self.images = images # [#img, C=12, H, W]
        self.labels = labels # [#gt/tile, C=2, H, W]
        self.locations = locations # [[tid, i, j]]
        self.valid_image_masks = valid_image_masks
        
        if tile_info is None:
            self.split_year = False
        else:
            self.split_year = True
            self.tiles, self.num_img_tile, self.idx_start_tile = tile_info[0], tile_info[1], tile_info[2]

        print('locations len: {}'.format(len(self.locations)))
        print('valid image masks shape: {}'.format(self.valid_image_masks.shape))

        self.patch_size = patch_size
        self.num_image_per_tile = int(len(images)/len(labels))

        print('num_image_per_tile: {}'.format(self.num_image_per_tile))

        self.image_transform = image_transform
        self.label_transform = label_transform

    def __getitem__(self, idx):
        tid, i, j = self.locations[idx]
        
        idx_start = self.idx_start_tile[self.tiles[tid]] if self.split_year else tid*self.num_image_per_tile
        idx_end = idx_start + self.num_img_tile[self.tiles[tid]] if self.split_year else (tid+1)*self.num_image_per_tile
        # valid image ids range in [1, num_image_per_tile+1], 0 means this image not valid
        valid_image_ids = self.valid_image_masks[idx_start:idx_end, i, j]
        #print('valid image ids (before 0 cond) {}'.format(valid_image_ids))
        valid_image_ids = valid_image_ids[valid_image_ids != 0]
        #print('valid image ids {}'.format(valid_image_ids))
        imgid = random.choice(valid_image_ids) - 1

        # get patch
        i_slice = slice(i - (self.patch_size // 2), i + (self.patch_size // 2) + 1)
        j_slice = slice(j - (self.patch_size // 2), j + (self.patch_size // 2) + 1)

        image_patch = self.image_transform(self.images[imgid+tid*self.num_image_per_tile, :, i_slice, j_slice])
        label_patch = self.label_transform(self.labels[tid, :, i_slice, j_slice])

        return image_patch, label_patch

    def __len__(self):
        return len(self.locations)

class RandomSampler(Sampler):
    """
    https://discuss.pytorch.org/t/new-subset-every-epoch/85018
    """

    def __init__(self, data_source, num_samples=None):
        super().__init__(data_source)
        self.data_source = data_source
        self._num_samples = num_samples

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError(
                "num_samples should be a positive integer "
                "value, but got num_samples={}".format(self.num_samples)
            )

    @property
    def num_samples(self):
        # dataset size might change at runtime
        if self._num_samples is None:
            return len(self.data_source)
        return self._num_samples

    def __iter__(self):
        n = len(self.data_source)
        return iter(torch.randperm(n, dtype=torch.int64)[:self.num_samples].tolist())

    def __len__(self):
        return self.num_samples

class ToTensor(torch.nn.Module):
    """
    In contrast to torchvision.transforms.ToTensor, this class doesn't permute the images' dimensions.
    """
    def forward(self, array):
        return torch.from_numpy(array).float()
